package com.me.pacman.state;

import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.Input;
import com.badlogic.gdx.InputAdapter;
import com.badlogic.gdx.graphics.Color;
import com.badlogic.gdx.graphics.Texture;
import com.badlogic.gdx.graphics.g2d.BitmapFont;
import com.badlogic.gdx.graphics.g2d.TextureRegion;
import com.badlogic.gdx.math.Vector2;
import com.me.pacman.PacDude;
import com.me.pacman.entity.*;
import com.me.pacman.entity.ai.ReturnToBase;
import com.me.pacman.level.Level;
import com.me.pacman.level.LevelTile;

import java.util.Random;

public class PlayState extends LevelState {

    public static final float FULL_SPEED = 9.375f;

    public static final Vector2[] GHOST_SPAWN_POINTS = {
        new Vector2(14f, 19.5f),
        new Vector2(14f, 16.5f),
        new Vector2(12f, 16.5f),
        new Vector2(16f, 16.5f),
    };

    public static final Direction[] GHOST_SPAWN_DIRS = {
            Direction.LEFT,
            Direction.UP,
            Direction.DOWN,
            Direction.DOWN,
    };

    private Texture levelBackground, winBackground;
    private TextureRegion lifeSprite;
    private BitmapFont font;

    public Random random;

    private long sirenId;

    private int pelletCount;
    private int pelletsEaten;

    public int pelletsEatenSinceDeath;
    public boolean pelletsEatenSinceDeathCounterEnabled;

    private int score;
    private int lives;
    private int round;
    private boolean paused = false;

    private float modeTimer;
    private GameState mode;

    public float secondsSinceLastDot;

    public boolean scatter;
    private int scatterCount;
    private float scatterChaseTimer;

    public float frightTimer;           // remaining fright time
    private int ghostsCaught;           // number of ghosts caught since last power pellet
    private Ghost lastGhostCaptured;

    public Pacman pacman;
    public Ghost[] ghosts;

    public PlayState(PacDude game) {
        super(game);
    }

    @Override
    public void setup() {
        levelBackground = game.assets.getLevelBackground();
        winBackground = game.assets.getLevelWinBackground();
        font = game.assets.getFont();
        lifeSprite = game.assets.pacman[2][1];
        Gdx.input.setInputProcessor(new Controller());

        preNewGame();
        setGameState(GameState.PRE_NEW_GAME);
    }

    @Override
    public void render() {
        level.render(0, 16);

        game.assets.getFont().setColor(Color.WHITE);
        game.assets.getFont().draw(game.batch, "" + score, 40, 279);
        for (int i = 0; i < lives; i++) {
            game.batch.draw(lifeSprite, i * 16, 0);
        }

        if (mode == GameState.GHOST_CAUGHT_POINTS_WAIT) {
            game.batch.draw(game.assets.points[0][ghostsCaught-1], (lastGhostCaptured.pos.x * 8) - 8, (lastGhostCaptured.pos.y * 8) + 8);
        } else {
            pacman.render(game.batch, 0, 16);
        }

        if (mode == GameState.ROUND_WON) {
            // draw flashing level background
            game.batch.draw((int) (modeTimer * 4) % 2 == 0? levelBackground : winBackground, 0, 16);
            return;
        } else {
            game.batch.draw(levelBackground, 0, 16);
        }

        switch (mode) {
            case PRE_NEW_GAME:
            case PACMAN_CAUGHT:
            case ROUND_WON:
            case GAME_OVER:
                break;
            default:
                for (Ghost ghost : ghosts) {
                    if (mode == GameState.GHOST_CAUGHT_POINTS_WAIT && ghost == lastGhostCaptured) {
                        game.batch.draw(game.assets.points[0][ghostsCaught - 1], (ghost.pos.x * 8) - 8, (ghost.pos.y * 8) + 8);
                    } else {
                        ghost.render(game.batch, 0, 16);
                    }
                }
        }

        if (paused) {
            game.assets.getFont().setColor(Color.YELLOW);
            game.assets.getFont().draw(game.batch, "paused", 90, 151);
        } else {
            switch (mode) {
                case PRE_NEW_GAME:
                case NEW_ROUND_WAIT:
                case START_ROUND_WAIT:
                    game.assets.getFont().setColor(Color.YELLOW);
                    game.assets.getFont().draw(game.batch, "ready!", 92, 127);
                    break;
                case GAME_OVER:
                    game.assets.getFont().setColor(Color.RED);
                    game.assets.getFont().draw(game.batch, "game  over", 78, 127);
                    break;
            }
        }
    }

    @Override
    public void update(float dt) {
        if (paused) {
            return;
        }

        if (PacDude.DEBUG) {
            // Fixed time step for debugger
            dt = 1/60f;
        }

        switch (mode) {
            case PLAYING:
                if (frightTimer <= 0) {
                    if (scatter || scatterCount < 4) {
                        scatterChaseTimer -= dt;
                        if (scatterChaseTimer <= 0) {
                            scatter = !scatter;
                            scatterChaseTimer = scatter ? 7f : 20f;
                            if (scatter) {
                                scatterCount++;
                            }
                        }
                    }
                } else {
                    frightTimer -= dt;
                    if (frightTimer <= 0) {
                        for (Ghost ghost : ghosts) {
                            if (ghost.currentBehaviour instanceof ReturnToBase) continue;
                            ghost.caught = false;
                            ghost.currentBehaviour = ghost.chaseBehaviour;
                        }
                    }
                }

                secondsSinceLastDot += dt;
                if (secondsSinceLastDot >= 4) {
                    // It's been 4 seconds since pacman last ate a dot, he's tryin' to avoid ghosts coming out!
                    // We'll get him...
                    for (int i = 1; i < 4; i++) {
                        Ghost ghost = ghosts[i];
                        if (ghost.inHouse) {
                            ghost.leaveHouse();
                            secondsSinceLastDot = 0;
                            break;
                        }
                    }
                }

                pacman.update(dt);
                for (Ghost ghost : ghosts) {
                    ghost.update(dt);

                    if (ghost.onSameTile(pacman)) {
                        if (frightTimer > 0 && !ghost.caught) {
                            ghostCaught(ghost);
                            continue;
                        }
                        if (!(ghost.currentBehaviour instanceof ReturnToBase)) {
                            pacmanCaught();
                            return;
                        }
                    }
                }
                break;
            case PACMAN_CAUGHT:
                pacman.update(dt);
                if (pacman.deathFrame == 13) {
                    if (lives > 0) {
                        resetField();
                        setGameState(GameState.START_ROUND_WAIT);
                    } else {
                        setGameState(GameState.GAME_OVER);
                    }
                }
                break;
            default:
                modeTimer -= dt;
                if (modeTimer <= 0) {
                    modeTransition();
                }
                break;
        }
    }

    public void modeTransition() {
        switch (mode) {
            // The mode we're transitioning /from/
            case PRE_NEW_GAME:
                newGame();
                setGameState(GameState.START_ROUND_WAIT);
                break;
            case START_ROUND_WAIT:
            case NEW_ROUND_WAIT:
                startGame();
                setGameState(GameState.PLAYING);
                break;
            case ROUND_WON_WAIT:
                setGameState(GameState.ROUND_WON);
                break;
            case ROUND_WON:
                newRound();
                setGameState(GameState.NEW_ROUND_WAIT);
                break;
            case GAME_OVER:
                preNewGame();
                setGameState(GameState.PRE_NEW_GAME);
                break;
            case GHOST_CAUGHT_POINTS_WAIT:
                setGameState(GameState.PLAYING);
                break;
            case PACMAN_CAUGHT_WAIT:
                game.assets.deathSound.play(1f);
                pacman.alive = false;
                setGameState(GameState.PACMAN_CAUGHT);
                break;
        }
    }

    public void initializeLevel() {
        level = new Level(game,"level");
        pelletCount = level.getTileCount(LevelTile.PELLET) + level.getTileCount(LevelTile.POWER_PELLET);
        pacman = new Pacman(this, false);
    }

    public void initializeRound() {
        scatter = true;
        scatterCount = 1;

        pelletsEaten = 0;
        pelletsEatenSinceDeath = 0;
        pelletsEatenSinceDeathCounterEnabled = false;

        spawnGhosts();
    }

    public void resetField() {
        lives--;
        frightTimer = 0f;
        ghostsCaught = 0;
        lastGhostCaptured = null;
        secondsSinceLastDot = 0;

        scatterChaseTimer = scatter? 7f : 20f;

        random = new Random(897198256012865L);

        pacman = new Pacman(this, false);
        resetGhosts();
    }

    public void preNewGame() {
        score = 0;
        lives = 3;
        round = 1;

        initializeLevel();

        game.assets.beginning.play(1.0f);
    }

    public void newGame() {
        initializeRound();
        resetField();
    }

    public void newRound() {
        initializeLevel();
        initializeRound();
        resetField();

        round++;

        game.assets.siren.stop(sirenId);
        game.assets.beginning_alt.play(1.0f);
    }

    public void startGame() {
        sirenId = game.assets.siren.loop(1.0f);
        pacman.moving = true;
    }

    private void spawnGhosts() {
        ghosts = new Ghost[4];
        ghosts[0] = new Blinky(this, new Vector2(GHOST_SPAWN_POINTS[0]), GHOST_SPAWN_DIRS[0]);
        ghosts[1] = new Pinky(this, new Vector2(GHOST_SPAWN_POINTS[1]), GHOST_SPAWN_DIRS[1]);
        ghosts[2] = new Inky(this, new Vector2(GHOST_SPAWN_POINTS[2]), GHOST_SPAWN_DIRS[2]);
        ghosts[3] = new Clyde(this, new Vector2(GHOST_SPAWN_POINTS[3]), GHOST_SPAWN_DIRS[3]);
    }

    public void resetGhosts() {
        for (int i = 0; i < 4; i++) {
            if (ghosts[i] == null) continue;
            ghosts[i].pos = new Vector2(GHOST_SPAWN_POINTS[i]);
            ghosts[i].currDirection = GHOST_SPAWN_DIRS[i];
            ghosts[i].currentPath = null;
            ghosts[i].caught = false;
            ghosts[i].inHouse = i > 0;
            ghosts[i].updateBehaviour();
        }
    }

    private void pelletEaten(float x, float y) {
        level.setTile(x, y, LevelTile.EMPTY);
        if (pelletsEaten % 2 == 0) {
            game.assets.chomp_1.play(1.0f);
        } else {
            game.assets.chomp_2.play(1.0f);
        }
        pelletsEaten++;
        pelletCount--;

        if (pelletsEatenSinceDeathCounterEnabled) {
            // Increase global dot counter when enabled
            pelletsEatenSinceDeath++;
        } else {
            // else increment appropriate ghost's individual counter
            for (Ghost ghost : ghosts) {
                // Increment the dot counter of the first encountered ghost in the house
                // Ghosts are ordered Blinky, Pinky, Inky, Clyde. Blinky's inHouse is always false,
                // Increase the dotCounter of only one ghost: the first of the above list still in the house.
                if (!ghost.inHouse) continue;
                ghost.dotCounter++;
                break;
            }
        }

        secondsSinceLastDot = 0;

        if (pelletCount == 0) {
            setGameState(GameState.ROUND_WON_WAIT);
        }
    }

    public void eatPellet(float x, float y) {
        pelletEaten(x, y);
        score += 10;
    }

    public void eatPowerPellet(float x, float y) {
        pelletEaten(x, y);
        for (Ghost ghost : ghosts) {
            if (ghost.currentBehaviour instanceof ReturnToBase) continue;
            ghost.caught = false;
            ghost.currentBehaviour = ghost.frightBehaviour;
            ghost.currDirection = ghost.currDirection.getOpposite();
        }
        frightTimer = 6f;
        ghostsCaught = 0;
        score += 50;
    }

    private void pacmanCaught() {
        game.assets.siren.stop(sirenId);
        pacman.moving = false;
        pelletsEatenSinceDeath = 0;
        pelletsEatenSinceDeathCounterEnabled = true;

        setGameState(GameState.PACMAN_CAUGHT_WAIT);
    }

    private void ghostCaught(Ghost ghost) {
        ghost.caught = true;
        ghost.currentBehaviour = new ReturnToBase(this);

        lastGhostCaptured = ghost;
        ghostsCaught++;
        score += ghostsCaught * 200;

        setGameState(GameState.GHOST_CAUGHT_POINTS_WAIT);
    }

    @Override
    public void dispose() {
        Gdx.input.setInputProcessor(null);
    }

    private final class Controller extends InputAdapter {

        @Override
        public boolean keyDown(int keycode) {
            switch (keycode) {
                case Input.Keys.UP:
                case Input.Keys.W:
                    pacman.setNextDirection(Direction.UP);
                    break;
                case Input.Keys.DOWN:
                case Input.Keys.S:
                    pacman.setNextDirection(Direction.DOWN);
                    break;
                case Input.Keys.LEFT:
                case Input.Keys.A:
                    pacman.setNextDirection(Direction.LEFT);
                    break;
                case Input.Keys.RIGHT:
                case Input.Keys.D:
                    pacman.setNextDirection(Direction.RIGHT);
                    break;
                case Input.Keys.ESCAPE:
                case Input.Keys.P:
                    paused = !paused;
                    break;
                case Input.Keys.N:
                    preNewGame();
                    break;
            }
            return super.keyDown(keycode);
        }

    }

    public enum GameState {

        PRE_NEW_GAME(2.3f),
        NEW_ROUND_WAIT(4.2f),
        START_ROUND_WAIT(2f),
        ROUND_WON_WAIT(1f),
        ROUND_WON(2f),
        PACMAN_CAUGHT,
        PACMAN_CAUGHT_WAIT(1f),
        GHOST_CAUGHT_POINTS_WAIT(1f),
        GAME_OVER(2f),
        PLAYING,
        ;

        final float timer;

        GameState(float timer) {
            this.timer = timer;
        }

        GameState() {
            this(0f);
        }

    }

    public void setGameState(GameState mode) {
        this.mode = mode;
        this.modeTimer = mode.timer;
    }

}
